/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the "Elastic License
 * 2.0", the "GNU Affero General Public License v3.0 only", and the "Server Side
 * Public License v 1"; you may not use this file except in compliance with, at
 * your election, the "Elastic License 2.0", the "GNU Affero General Public
 * License v3.0 only", or the "Server Side Public License, v 1".
 */

import { nodeTypes } from '../node_types';
import { fields } from '../../filters/stubs';
import { DataViewBase } from '../../..';

import * as range from './range';
import * as estypes from '@elastic/elasticsearch/lib/api/typesWithBodyKey';
import { KQL_NODE_TYPE_LITERAL } from '../node_types/literal';
import { KqlRangeFunctionNode } from './range';

describe('kuery functions', () => {
  describe('range', () => {
    let indexPattern: DataViewBase;

    beforeEach(() => {
      indexPattern = {
        fields,
        title: 'dataView',
      };
    });

    describe('buildNodeParams', () => {
      test('arguments should contain the provided fieldName as a literal', () => {
        const result = range.buildNodeParams('bytes', 'gt', 1000);
        const {
          arguments: [fieldName],
        } = result;

        expect(fieldName).toHaveProperty('type', KQL_NODE_TYPE_LITERAL);
        expect(fieldName).toHaveProperty('value', 'bytes');
      });

      test('arguments should contain the provided value as a literal', () => {
        const result = range.buildNodeParams('bytes', 'gt', 1000);
        const {
          arguments: [, , valueArg],
        } = result;

        expect(valueArg).toHaveProperty('type', KQL_NODE_TYPE_LITERAL);
        expect(valueArg).toHaveProperty('value', 1000);
      });
    });

    describe('toElasticsearchQuery', () => {
      test("should return an ES range query for the node's field and params", () => {
        const expected = {
          bool: {
            should: [
              {
                range: {
                  bytes: {
                    gt: 1000,
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          'bytes',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern);

        expect(result).toEqual(expected);
      });

      test('should return an ES range query without an index pattern', () => {
        const expected = {
          bool: {
            should: [
              {
                range: {
                  bytes: {
                    gt: 1000,
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };

        const node = nodeTypes.function.buildNode(
          'range',
          'bytes',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node);

        expect(result).toEqual(expected);
      });

      test('should support wildcard field names', () => {
        const expected = {
          bool: {
            should: [
              {
                range: {
                  bytes: {
                    gt: 1000,
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };

        const node = nodeTypes.function.buildNode(
          'range',
          'byt*',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern);

        expect(result).toEqual(expected);
      });

      test('should support scripted fields', () => {
        const node = nodeTypes.function.buildNode(
          'range',
          'script number',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern);

        expect((result.bool!.should as estypes.QueryDslQueryContainer[])[0]).toHaveProperty(
          'script'
        );
      });

      test('should support date fields without a dateFormat provided', () => {
        const expected = {
          bool: {
            should: [
              {
                range: {
                  '@timestamp': {
                    gt: '2018-01-03T19:04:17',
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          '@timestamp',
          'gt',
          '2018-01-03T19:04:17'
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern);

        expect(result).toEqual(expected);
      });

      test('should support date fields with a dateFormat provided', () => {
        const config = { dateFormatTZ: 'America/Phoenix' };
        const expected = {
          bool: {
            should: [
              {
                range: {
                  '@timestamp': {
                    gt: '2018-01-03T19:04:17',
                    time_zone: 'America/Phoenix',
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          '@timestamp',
          'gt',
          '2018-01-03T19:04:17'
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern, config);

        expect(result).toEqual(expected);
      });

      test('should use a provided nested context to create a full field name', () => {
        const expected = {
          bool: {
            should: [
              {
                range: {
                  'nestedField.bytes': {
                    gt: 1000,
                  },
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          'bytes',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(
          node,
          indexPattern,
          {},
          { nested: { path: 'nestedField' } }
        );

        expect(result).toEqual(expected);
      });

      test('should automatically add a nested query when a wildcard field name covers a nested field', () => {
        const expected = {
          bool: {
            should: [
              {
                nested: {
                  path: 'nestedField.nestedChild',
                  query: {
                    range: {
                      'nestedField.nestedChild.doublyNestedChild': {
                        lt: 8000,
                      },
                    },
                  },
                  score_mode: 'none',
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          '*doublyNested*',
          'lt',
          8000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern);

        expect(result).toEqual(expected);
      });

      test('should allow to configure ignore_unmapped for a nested query', () => {
        const expected = {
          bool: {
            should: [
              {
                nested: {
                  path: 'nestedField.nestedChild',
                  query: {
                    range: {
                      'nestedField.nestedChild.doublyNestedChild': {
                        lt: 8000,
                      },
                    },
                  },
                  score_mode: 'none',
                  ignore_unmapped: true,
                },
              },
            ],
            minimum_should_match: 1,
          },
        };
        const node = nodeTypes.function.buildNode(
          'range',
          '*doublyNested*',
          'lt',
          8000
        ) as KqlRangeFunctionNode;
        const result = range.toElasticsearchQuery(node, indexPattern, {
          nestedIgnoreUnmapped: true,
        });

        expect(result).toEqual(expected);
      });
    });

    describe('toKqlExpression', () => {
      describe('operators', () => {
        test('gt', () => {
          const node = nodeTypes.function.buildNode(
            'range',
            'bytes',
            'gt',
            1000
          ) as KqlRangeFunctionNode;
          const result = range.toKqlExpression(node);
          expect(result).toMatchInlineSnapshot(`"bytes > 1000"`);
        });

        test('gte', () => {
          const node = nodeTypes.function.buildNode(
            'range',
            'bytes',
            'gte',
            1000
          ) as KqlRangeFunctionNode;
          const result = range.toKqlExpression(node);
          expect(result).toMatchInlineSnapshot(`"bytes >= 1000"`);
        });

        test('lt', () => {
          const node = nodeTypes.function.buildNode(
            'range',
            'bytes',
            'lt',
            1000
          ) as KqlRangeFunctionNode;
          const result = range.toKqlExpression(node);
          expect(result).toMatchInlineSnapshot(`"bytes < 1000"`);
        });

        test('lte', () => {
          const node = nodeTypes.function.buildNode(
            'range',
            'bytes',
            'lte',
            1000
          ) as KqlRangeFunctionNode;
          const result = range.toKqlExpression(node);
          expect(result).toMatchInlineSnapshot(`"bytes <= 1000"`);
        });
      });

      test('with wildcard field & literal value', () => {
        const node = nodeTypes.function.buildNode(
          'range',
          'byt*',
          'gt',
          1000
        ) as KqlRangeFunctionNode;
        const result = range.toKqlExpression(node);
        expect(result).toMatchInlineSnapshot(`"byt* > 1000"`);
      });

      test('with wildcard field & value', () => {
        const node = nodeTypes.function.buildNode(
          'range',
          'byt*',
          'gt',
          '100*'
        ) as KqlRangeFunctionNode;
        const result = range.toKqlExpression(node);
        expect(result).toMatchInlineSnapshot(`"byt* > 100\\\\*"`);
      });
    });
  });
});
